/*
 * BMOutputStream.java
 *
 * Created on 25.12. 2003, 20:48
 */

package saiko.compression;

import java.io.ByteArrayOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * Class to provide simple compression algorithm output stream
 * using bit maps - replacing the occurence of the most often used character
 * by bit maps of this occurence positions.
 * be aware - flush() closes the stream.
 *
 * @author  Dusan Saiko
 * @see saiko.compression.BMInputStream
 */
public class BitMapsOutputStream extends FilterOutputStream {
    
    /**
     * the header sign on the compressed data
     * DSCBM1 - Duan Saiko Compression Bit Maps version 1
     */
    public final static String header="DSCBM1";
    
    private int theChar;    //the most often occured character
    private ByteArrayOutputStream buffer;  //data buffer for analyzing the data stream for 
                                           //the most occured character
    private int  bufferSize; //max size of the data in the buffer
    
    /** 
     * Constructor of new instance of BMOutputStream
     * Constructor sets the default buffer size to be 16 MB.
     *
     * @param aOutput (OutputStream) - enclosed output stream
     * @exception  IOException if an I/O error occurs
    */
    public BitMapsOutputStream(OutputStream aOutput) throws IOException
    {
        this(aOutput, 16*1024*1024);
    }

    /** 
     * Constructor of new instance of BMOutputStream
     *
     * @param aOutput (OutputStream) - enclosed output stream
     * @param aBufferSize (int)     -  size of the buffer to analyze for the
     *          most often used character, if set to 0, than it is expected 
     *          you set the value of the most often used character wich will
     *          be compressed by calling the setCompressChar
     * @exception  IOException if an I/O error occurs
    */
    public BitMapsOutputStream(OutputStream aOutput, int aBufferSize) throws IOException
    {
        super(aOutput);
        theChar = -1; //initialize the current char to some out of range value
        out.write(header.getBytes());
        buffer=new ByteArrayOutputStream();
        bufferSize=Math.max(64,aBufferSize);
    }

    /**
     * sets the character which is compressed
     * if this function is not called, the most often occured character is
     * search in the data automatically
     *
     * @param aChar (int) - most often occured character in the data
     */
    public synchronized void setCompressChar(int aChar)
    throws IllegalArgumentException, IOException
    {
        if(aChar<0) throw new IllegalArgumentException();
        if(theChar!=-1) throw new IllegalStateException("Compression character already set.");
        if(theChar==-1)
            out.write(aChar);
        theChar=aChar;
    }
    
    /**
     * method analyzes buffer and sets the most often occured character
     */
    private synchronized void analyzeBuffer() 
    throws IOException
    {
        if(buffer==null) return;
        byte data[]=buffer.toByteArray();
        if(data==null) return;
        
        int counts[]=new int[256];  //count the counts of each character
        int i, max=0;

        for(i=0; i<counts.length; i++) counts[i]=0;
        
        //go through the data and count the characters
        for(i=0; i<data.length; i++) {
            int value = ((int)(char)data[i])&0xFF;
            counts[value]++;
        }
        
        //search for most often used character
        for(i=0; i<counts.length; i++) {
            if(counts[i]>counts[max]) max=i;
        }
        data=null;
        counts=null;
        setCompressChar(max);
    }
    
    /**
     * Flush the remaining data to the output
     */
    private synchronized void flushBuffer() throws IOException {
        if(theChar==-1) {
            analyzeBuffer();
        }
        byte data[]=buffer.toByteArray();
        
        int pos=0;
        //for all the data in buffer
        while(pos<=data.length-7) {
            int value=((int)(char)data[pos])&0xFF;
            pos++;

            //if the current character is not the most often used, just output it
            if((value!=theChar) && ((value&0x80)==0)) {
                out.write(value);
                //loop again
                continue;
            } 
            
            //check if there is still at least 6 more bytes available at the 
            // data - we need it for the map
            if(pos>data.length-6) break; 

            //create the map
            
            int i, map;
            
            // we have read the most often used character, or the character with
            // first bit set
            if(value==theChar) {
                // 10111111 => first bit means this is the map, the second
                // the most often used character is next
                map=0xBF;
            } else {
                // 11111111 => first bit means this is the map, the second
                // the next character will be in the output
                map=0xFF;
            }
            
            // read next 6 values
            int vals[]=new int[6];
            for(i=0; i<vals.length; i++) {
                vals[i]=((int)(char)data[pos])&0xFF;
                pos++;
            }
            // go through values and clear the bit in map on its position
            // if it is the most often used character
            for(i=0; i<vals.length; i++) {
             if(vals[i]==theChar) {
                 map=map&~(1<<(5-i)); //clear the bit in map
             }
            } //for
            
            //output the map and the data
            out.write(map);
            
            //if we are creating the map just because of the character
            //which has first bit set, we have to output this byte
            //map already indicates it
            if(value!=theChar) {
                out.write(value);
            }
            
            //output the characters which differs from the most used one
            for(i=0; i<vals.length; i++) {
             if(vals[i]!=theChar) {
                 out.write(vals[i]);
             }
            } //for
            
        } //while
        
        // clear the buffer
        buffer.reset();
        
        //copy  remaining data into the new buffer
        while(pos<data.length) {
            int value=((int)(char)data[pos])&0xFF;
            buffer.write(value);
            pos++;
        }
        data=null;
    }

    /**
     * Writes the specified byte to this compressed output stream. 
     *
     * @param      aValue (int)  the byte to be written.
     * @exception  IOException  if an I/O error occurs.
     */
    public synchronized void write(int aValue) throws IOException {
        buffer.write(aValue);
        if(buffer.size()>=bufferSize) {
            flushBuffer();
        }
    }

  
    /**
     * Flushes this buffered output stream. This forces any buffered 
     * output bytes to be written out to the underlying output stream. 
     * because of algorithm data cashing, flush closes the stream
     *
     * @exception  IOException  if an I/O error occurs.
     * @see        java.io.FilterOutputStream#out
     */
    public synchronized void flush() throws IOException {
        flushBuffer();
        //output rest from the buffer
        byte data[]=buffer.toByteArray();
        if(data!=null && data.length>0) {
            out.write(0xFF); 
            //map - write everything behind it (in case there is any
            // character with first bit set
            for(int i=0; i<data.length; i++) {
                int value=((int)(char)data[i])&0xFF;
                out.write(value);
            }
        }
	out.flush();
        out.close();
    }     
}
