/*
 * SourceStatistics.java
 *
 * Created on 4.1. 2004, 17:46
 */

package saiko.compression;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;

/**
 * class for getting the statistics for data source
 *
 * @author  Dusan Saiko
 */
public class SourceStatistics {
    
    private byte[] data;

    /** 
     * Constructor of new instance of SourceStatistics
     * Statistics are computed on some data
     *
     * @param aData (byte[]) - data itself
    */    
    public SourceStatistics(byte[] aData)
    {
        data=aData;
    }

    /** 
     * Helper constructor of new instance of SourceStatistics
     * where data are read directly from a file
     *
     * @param aFile (File) - data file
    */    
    public SourceStatistics(File aFile)
    throws IOException
    {
        if(aFile==null) return;
        ByteArrayOutputStream buffer=new ByteArrayOutputStream((int)aFile.length());
        byte b[]=new byte[1024*16];
        int size;
        FileInputStream input=new FileInputStream(aFile);
        while((size=input.read(b))>=0) {
            buffer.write(b,0,size);
        }
        data=buffer.toByteArray();
        input.close();
    }
    
    /**
     * get statistics of aOrder order
     * @param aOrder - 
     * @return Statistics[]
     * @see saiko.compression.Statistic
     */
    private Statistics[] getStats(int aOrder) 
    {
        HashMap table=new HashMap(1024);

        int order=aOrder;
        if(order==0) order=1;
        
        int buffer[]=new int[order];
        
        // create the buffer and initialize it to "null" values -1
        for(int i=0; i<buffer.length; i++) buffer[i]=-1;
        
        // for all the data
        for(int i=0; i<data.length; i++) {
            //if there are null values, fill the buffer
            boolean filled=true;
            boolean inserted=false;
            
            for(int n=0; n<buffer.length; n++) {
                if(buffer[n]==-1) {
                    buffer[n]=data[i];
                    if(n!=buffer.length-1) filled=false;
                    inserted=true;
                    break;
                }
            }
            if(filled) {
                if(!inserted) {
                    //shift the buffer 
                    for(int n=0; n<buffer.length-1; n++) {
                        buffer[n]=buffer[n+1];
                    }
                    buffer[buffer.length-1]=(((int)(char)data[i])&0xFF);
                }

                //increment count and place it into stats
                Statistics sample=new Statistics(buffer);
                BigInteger hashCode=sample.hash();
                Statistics stats=(Statistics) table.get(hashCode);
                if(stats==null) {
                     table.put(hashCode,sample);
                } else {
                    stats.count++;
                }
            }
        }
        
        Statistics[] res=null;
        res=(Statistics[]) table.values().toArray(new Statistics[]{});
        return res;
    }
    
  /**
     * get statistics of aOrder order
     * @param aOrder - 
     * @return Statistics[]
     * @see saiko.compression.Statistic
     */
    private Statistics[] getAlignedStats(int aOrder) 
    {
        HashMap table=new HashMap(1024);

        int order=aOrder;
        if(order==0) order=1;
        
        int buffer[]=new int[order];
        
        // create the buffer and initialize it to "null" values -1
        for(int i=0; i<buffer.length; i++) buffer[i]=-1;
        
        // for all the data
        int i=0;
        while(i<data.length) {
            //if there are null values, fill the buffer
            boolean filled=true;
            
            for(int n=0; n<buffer.length; n++) {
                if(buffer[n]==-1) {
                    buffer[n]=data[i];
                    if(n!=buffer.length-1) filled=false;
                    break;
                }
            }
            if(filled) {
                //increment count and place it into stats
                Statistics sample=new Statistics(buffer);
                BigInteger hashCode=sample.hash();
                Statistics stats=(Statistics) table.get(hashCode);
                if(stats==null) {
                     table.put(hashCode,sample);
                } else {
                    stats.count++;
                }
                
                //clear the buffer 
                for(int n=0; n<buffer.length; n++) {
                    buffer[n]=-1;
                }
                
            }
            
            i++;
        } //while
        
        Statistics[] res=null;
        res=(Statistics[]) table.values().toArray(new Statistics[]{});
        return res;
    }
    
    /**
     * sorts the statistics according to DESC order
     * @param aStats (Statistics[]) - statistics to be sorted
     */ 
    public static void sortStats(Statistics[] aStats) {
        Arrays.sort(aStats, new Comparator() {
                public int compare(Object o1, Object o2) {
                    Statistics s1=(Statistics) o1;
                    Statistics s2=(Statistics) o2;
                    if(s1.count>s2.count) return -1;
                    if(s2.count>s1.count) return 1;
                    for(int i=0; i<Math.min(s1.data.length, s2.data.length); i++)
                    {
                        if(s1.data[i]<s2.data[i]) return -1;
                        if(s1.data[i]>s2.data[i]) return 1;
                    }
                    if(s1.data.length<s2.data.length) return -1;
                    if(s1.data.length>s2.data.length) return 1;
                    return 0;
                }
            });
    }
    
    /**
     * returns statistics on the data for aOrder order model
     * @param aOrder (int) - Shannon order model number
     * @param aSortDescending (boolean) - flag for sorting the output according
     *        to number of occurances in descending order
     * @return Statistics - resulting statistics or null if data length is smaller than aOrder
     */
    public Statistics[] getStats(int aOrder, boolean aSortDescending) 
    throws IllegalArgumentException
    {
        if(aOrder<0) throw new IllegalArgumentException("Order can not be less tnan 0");
        if(aOrder>data.length) return null;
        
        Statistics[] res=null;

        res=getStats(aOrder);

        //if order 0 was specified, then no character counts are needed
        if(aOrder==0)
            for(int i=0; i<res.length; i++) res[i].count=1;
        
        //sort if wanted
        if(aSortDescending && res!=null) {
            sortStats(res);
        }
        
        return res;
    }

    /**
     * returns statistics on the data for aOrder order model
     * the stats gets only on data alignet to the number of aOrder bytes 
     * this means, for CBAABA gets only statistics CB AA BA, not the full
     * statistics CB BA AA AB BA
     * @param aOrder (int) - Shannon order model number
     * @param aSortDescending (boolean) - flag for sorting the output according
     *        to number of occurances in descending order
     * @return Statistics - resulting statistics or null if data length is smaller than aOrder
     */
    public Statistics[] getAlignedStats(int aOrder, boolean aSortDescending) 
    throws IllegalArgumentException
    {
        if(aOrder<0) throw new IllegalArgumentException("Order can not be less tnan 0");
        if(aOrder>data.length) return null;
        
        Statistics[] res=null;

        res=getAlignedStats(aOrder);

        //if order 0 was specified, then no character counts are needed
        if(aOrder==0)
            for(int i=0; i<res.length; i++) res[i].count=1;
        
        //sort if wanted
        if(aSortDescending && res!=null) {
            sortStats(res);
        }
        
        return res;
    }
}
